#!/usr/bin/python
#
# NamedConf.py : named.conf parser
#
# Jason Vas Dias <jvdias@redhat.com> Aug 2004
#
# A "Partially Correct" Parse of named.conf is not enough!
#
import sys
import os
import string
import commands
import datetime
import md5

class NamedConf:
    
    def directory(self):
        if( ( not self.objs.has_key('options') ) or ( not self.objs['options'].has_key('directory')) ):
            return self.chroot + '/var/named' 
        dir = self.objs['options']['directory'].strip('"')
        if dir[-1] == '/':
            dir = dir[:-1]
        if dir[0] != '/':
            dir = '/' + dir
        return self.chroot + dir
    
    def zone_origin_class(self, k):                        
        zk=k.split()
        if len(zk) > 1:
            return [ zk[0].strip('"') , zk[1].upper() ]
        else:
            return [ zk[0].strip('"') , "IN" ]

    def zone_origin(self, k):
        return self.zone_origin_class(k)[0]

    def zone_class(self,k):
        return self.zone_origin_class(k)[1]

    def zone_file(self,k,view=None):
        zone=None
        if view != None:
            zone=self.objs['view'][view]['zone'][k]
        else:
            zone=self.objs['zone'][k]
        if zone.has_key('file'):
            file=zone['file']
            file=file.strip('"')
            if file[0] != '/':
                file=self.directory()+'/'+file
            return file
        return None
    
    def zones(self):
        objs=self.objs
        if objs.has_key('view'):
            # then ALL zones must be in a view (named.conf syntax)
            for v in objs['view'].keys():
                if objs['view'][v].has_key('zone'):
                   for z in  objs['view'][v]['zone']:
                       yield([self.zone_origin(z), z, self.zone_file(z,v), self.zone_class(z), objs['view'][v]['zone'][z], v])
        elif objs.has_key('zone'):            
            for z in objs['zone'].keys():
                yield([self.zone_origin(z), z, self.zone_file(z), self.zone_class(z), objs['zone'][z], None])
    
    def parse( self ):
        inLineComment=0
        inMultiLineComment=0
        inString=0
        token=''
        tokens=[]
        keys=[]
        keyCounts=[]
        p=''
        end=0
        begin=0
        first_tok=0
        last_tok=0
        obj=self.objs
        id=''
        first_tok=0
        controlKey=None
        for c in self.contents[:]:            
            if (inString==0) and (inLineComment==0) and (inMultiLineComment==0) and (((c=='/') and (p=='/')) or ( (c == '#') and (p != '\\' ))):
               inLineComment=1
               if c == '/':
                   token=token[:-1]
               if len(token) > 0:
                   tokens.append(token)
               begin=end
               end+=1
               token=''
               continue
            if (inLineComment==1) and (c=='\n') and (p != '\\'):
               inLineComment = 0
               begin=end
               end+=1
               token=''
               continue
            if (inString==0) and (inMultiLineComment==0) and (inLineComment==0) and (p=='/') and (c=='*'):
               inMultiLineComment=1
               if len(token)>1:
                   tokens.append(token[:-1])
               begin=end
               end+=1
               token=''
               continue
            if (inString==0) and (inMultiLineComment==1) and (inLineComment==0) and  (p=='*') and (c=='/'):
               inMultiLineComment=0
               begin=end
               end+=1
               token=''
               continue
            if (c=='"') and (p != '\\') and not(inMultiLineComment or inLineComment):
               inString=not inString            
            if (inMultiLineComment==0) and (inLineComment==0) and (inString==0) and (c in ['\n',' ','\t',';','{','}']):
               # end of non-comment not-in-string token:
               if len(token):
                   tokens.append(token)
                   self.tokens.append({'t': token, 'b': begin, 'e': end})
                   token=''
               if c == '{' :
                   if len(tokens) > 1:
                       key=string.join(tokens[1:],' ')
                       id=tokens[0].lower()
                   elif len(tokens):
                       key=tokens[0].lower()
                       id=''
                   else:
                       # embedded ACLs are legal (eg. 'acl test { {localhost; localnets}; };' )
                       key=''
                       id=''
                   obj=self.objs
                   for k in keys[:]:
                       obj=obj[k]
                   if len(id):
                       if not obj.has_key(id):
                           obj[id]={}
                       #print c, id, '->', key
                       if len(keys) and (keys[0] == 'controls'):
                           # allow controls to break one '{...}' per ';' rule
                           controlKey = [id,key]
                       else:
                           controlKey = None
                       obj[id][key]={}#{'_t':tokens}
                       obj=obj[id][key]
                       keys.append(id)
                       keys.append(key)
                       keyCounts.append(2)
                   else:
                       #print '{', key
                       if len(keys) and (keys[0] == 'controls'):
                           # allow controls to break one '{...}' per ';' rule
                           if controlKey != None:
                               ckc=0
                               for ck in controlKey:
                                   keys.append(ck)
                                   ckc+=1
                                   obj = obj[ck]
                               keyCounts.append(ckc+1)                               
#                          print 'CONTROLS NEW KEY:',key
                           keys.append(key)
                       else:
                           keys.append(key)
                           keyCounts.append(1)
                       if( type(obj) == list ):
                           # embedded ACLs only
                           obj.append({})
                           obj=obj[-1]
                       elif not obj.has_key(key):
                           obj[key]={}#{'_t':tokens}
                           obj=obj[key]
                       else:
                           # embedded ACLs only
                           if( type(obj[key]) != list ):
                               obj[key]= [ obj[key] ]
                           obj[key].append({})                       
                           obj=obj[key][-1]
                   for i in range((len(self.tokens) - len(tokens)), len(self.tokens)):                       
                       self.tokens[i]['keys']=keys[:]
                       self.tokens[i]['type']='key'
                   tokens=[]
                   id=''
               elif c == '}' :
                   i=keyCounts.pop()
                   while (i>0):
                       #print c, ':',keys.pop()
                       keys.pop()
                       i-=1
                   obj=self.objs
                   for k in keys[:]:
                       obj=obj[k]
                   tokens=[]
               elif (c == ';') and (len(tokens)):
                   joined=False
                   if len(tokens) > 1:
                       if len(tokens) > 2:                           
                           val=tokens[1:]                               
                       else:
                           val=tokens[1]                           
                           if (len(val)>1) and  (val[0] != '"') and (val[-1] != '"'):
                               val=val.lower()
                       if( len(tokens) > 1):
                           for t in tokens:
                               # ACLs only!
                               if (t[0] in '!/') or (t[-1] in '!/'):
                                   joined=True
                           if joined :
                               val=None
                       for i in range((len(self.tokens) - len(tokens)) + 1, len(self.tokens)):
                           self.tokens[i]['keys']=keys[:]
                           if not joined:
                               self.tokens[i]['keys'].append(tokens[0].lower())
                               self.tokens[i]['type']='val'
                           else:
                               self.tokens[i]['keys'].append(string.join(tokens,''))
                               self.tokens[i]['type']='key'
                       i=(len(self.tokens) - len(tokens))                           
                   else:
                       i=len(self.tokens)-1
                       val=None
                   k=tokens[0].lower()
                   if joined:
                       k = string.join(tokens,'').lower()
                       val=None
                   self.tokens[i]['keys']=keys[:]
                   self.tokens[i]['keys'].append(k)
                   self.tokens[i]['type']='key'
                   if( type(obj) == list):
                       # embedded ACLs only
                       obj=obj[-1]
                   if obj.has_key(k):
                       if len(keys) and (keys[0] == 'trusted-keys') and (type(obj[k])==list) and (type(val)==list):
                           if type(obj[k][0]) != list:
                               obj[k]=[obj[k]]
                           obj[k].append(val)
                       else:
                           if type(obj[k]) != list :
                               obj[k]=[obj[k]]
                           obj[k].append(val)
                       if len(tokens) > 1:
                           for i in range((len(self.tokens) - len(tokens)) + 1, len(self.tokens)):
                               self.tokens[i]['i']=len(obj[k])
                   else:
                       obj[k]=val
                   tokens=[]
               begin=end
            elif (inMultiLineComment==0) and (inLineComment==0):
               token += c            
            end+=1
            p=c
    
    def read(self, named_conf):
        try:
            fd = os.open(named_conf, os.O_RDONLY)
            s = os.fstat(fd)
            self.contents = os.read(fd, s.st_size)        
            os.close(fd)
        except OSError, e :
            return e
        self.parse()
        return 0

    def load(self, named_conf):
        s=commands.getstatusoutput('. /etc/sysconfig/named; echo "$ROOTDIR"')
        self.errors=[]
        self.files=[]
        self.objs={}
        self.lists={}
        self.noHeader=False
        opts=''
        if len(s[1]):
           self.chroot=s[1]
           if self.chroot[-1] == '/':
               self.chroot = self.chroot[:-1]
           opts+=' -t '+self.chroot
        if (named_conf != None) and len(named_conf):
           opts+=named_conf
           self.files.append({'file':named_conf,'tokens':[],'contents':''})
        else:
           self.files.append({'file':self.chroot + '/etc/named.conf','tokens':[],'contents':''})           
        s=commands.getstatusoutput('/usr/sbin/named-checkconf '+opts)
        if len(s[1]):
            self.errors.append(s[1])
            return s
        self.tokens=self.files[0]['tokens']
        self.contents=''
        self.read(self.files[0]['file'])
        if self.objs.has_key('include'):
            i=0
            n=0
            j=0
            while n < len( self.objs['include'] ):
                n = len( self.objs['include'] )
                includes=[]
                if type(self.objs['include']) == list:
                    includes = self.objs['include'][j:n]
                else:
                    includes = [self.objs['include']]
                for f in includes:
                    self.files[i]['contents']=self.contents
                    self.files[i]['md5']=md5.new(self.contents).digest()
                    self.contents=''
                    fn = f.strip('"')
                    if( fn[0] != '/'):
                        fn = '/' + fn
                    if len(self.chroot):
                        self.files.append({'file':self.chroot + fn ,'tokens':[],'contents':''})
                    else:
                        self.files.append({'file': fn, 'tokens':[],'contents':''})
                    i=i+1
                    self.tokens = self.files[i]['tokens']
                    e = self.read(self.files[i]['file'])
                    if e != 0:
                        self.errors=e[1]
                        return e
                j=n
            self.tokens=self.files[0]['tokens']
            self.files[i]['contents']=self.contents
            self.files[i]['md5']=md5.new(self.contents).digest()
            self.contents=self.files[0]['contents']
        else:
            self.files[0]['contents']=self.contents
            self.files[0]['md5'] = md5.new(self.contents).digest()
        return 0

    def nth_semicolon(s, nth, begin, end):
        i=begin
        inLineComment=0
        inMultiLineComment=0
        inString=0
        p=''
        semiColonIndex=-1
        n=0
        for c in s.contents[begin : end+1 ]:            
            if (inString==0) and (inLineComment==0) and (inMultiLineComment==0) and (((c=='/') and (p=='/')) or ((c == '#') and (p != '\\'))):
               inLineComment=1
               i += 1
               continue
            if (inLineComment==1) and (c=='\n'):
               inLineComment = 0
               i += 1
               continue
            if (inString==0) and (inMultiLineComment==0) and (inLineComment==0) and (p=='/') and (c=='*'):
               inMultiLineComment=1
               i += 1
               continue
            if (inString==0) and (inMultiLineComment==1) and (inLineComment==0) and  (p=='*') and (c=='/'):
               inMultiLineComment=0
               i += 1
               continue
            if (c=='"') and (p != '\\') and  not(inMultiLineComment or inLineComment):
               inString=not inString            
            if (inMultiLineComment==0) and (inLineComment==0) and (inString==0) and (c == ';'):
                n+=1
                semiColonIndex=i
                if (nth >= 0) and (n >= nth):
                    return semiColonIndex
            i+=1
            p=c
        return semiColonIndex
        
    def out(self, dumpKey=None):
        debug( 'NMDC OUT')
        if len(self.tokens) == 0:
            return self.contents
        last_key=''
        exists=True
        lastExists=True
        endOfKey=True
        output=''
        k_removed=-1
        parent={}
        dumpedKey = False
        if (dumpKey == None) and  (self.noHeader == False) and (self.tokens[0]['b'] > 0):
            output = self.contents[0:self.tokens[0]['b']+1]
        for i in range(0,len(self.tokens)):
            t = self.tokens[i]
            if (    (dumpKey != None) 
                and (   ( len(t['keys']) < len(dumpKey) )
                      or( t['keys'][:len(dumpKey)] != dumpKey)
                    )
               ):
                if dumpedKey:
                    break
                continue
            elif dumpKey != None:
                dumpedKey = True
            obj=self.objs
            lastExists=exists
            exists=True
            parent=obj
            kn=0
            for k in t['keys']:                
                if not obj.has_key(k):
                    kn+=1
                    exists=False
                    k_removed = kn
#                    print 'KEY REMOVED:', k_removed,t['keys'],k,kn
                else:
                    parent=obj
                    obj = obj[k]                
                if( type(obj) != dict ):
                    # embedded acls again!
                    break
            if exists and (not lastExists):
#                print 'K_REMOVED:',k_removed
                output+=self.contents[ self.nth_semicolon(k_removed, self.tokens[i-1]['e'], t['b'])+1 : t['b'] + 1 ]
                k_removed=0
            if parent.has_key('_NEW_') and (dumpKey == None):
                output+=string.join(parent['_NEW_'],'')
                del parent['_NEW_'] 
            if exists:
                if t['type']=='key':
                   output+=t['t']
                else:
                   if type(obj) == list:
                       lkey=string.join(t['keys'],'_')
                       if not self.lists.has_key(lkey):
                           if ( type(obj[0]) == list):
                               self.lists[lkey]=[0,0]
                           else:
                               self.lists[lkey]=0
                       li=0
                       lli=None
                       if type(self.lists[lkey]) == list:
                           li,lli=self.lists[lkey]
                       else:
                           li=self.lists[lkey]
                       if(len(obj)> li):
                           if ( type(obj[li]) == list):
                               if(len(obj[li])>lli):
                                   output += obj[li][lli]
                                   lli+=1
                               else:
                                   li+=1
                                   lli=0
                                   output += obj[li][lli]
                                   lli+=1
                               self.lists[lkey]=[li,lli]                               
                           else:
                               output+=obj[li]
                               self.lists[lkey]+=1
                       else:
                           del self.lists[lkey]
                   else:
                       if (i>0) and ((self.tokens[i-1]['type']=='key') or(self.tokens[i-1]['keys'] != t['keys'])):
                           if( t['t'].lower() == obj ):
                               output+=t['t']
                           else:
                               output+=obj
                if( (type(obj)==dict) and (obj.has_key('_NEW_'))) and (dumpKey == None) :
                    # handle empty "options" clause (the only {} exp. allowed to be empty!)
                    # which now has a new member
                    emptyBracket=1
                    for oK in obj.keys():
                        if( oK != '_NEW_' ):
                            emptyBracket=0                    
                    if emptyBracket:
                        emptyBracket = self.contents[t['e']: self.tokens[i+1]['b']+1 ].rfind('}')
                    if emptyBracket > 0:
                        output += self.contents[ t['e'] : t['e'] + emptyBracket ]
                        output += string.join(obj['_NEW_'],'')
                        del obj['_NEW_']
                        output += self.contents[ t['e'] + emptyBracket :  self.tokens[i+1]['b']+1 ]
                        continue
                if( (i+1) < len(self.tokens) ):
                    if( (dumpKey == None) or (    (len(  self.tokens[i+1]['keys'] ) >= len(dumpKey) )
                                              and (self.tokens[i+1]['keys'][:len(dumpKey)]==dumpKey)
                                             )
                      ):
                        output += self.contents[t['e']: self.tokens[i+1]['b']+1 ]
                    else:
                        output += self.contents[t['e']: self.nth_semicolon((len(self.tokens[i]['keys'])-len(dumpKey))+1, t['e'], self.tokens[i+1]['b']+1)+1 ]                       
                else:
                    output += self.contents[t['e']:]
            if ( len(self.tokens) == (i+1)) or (self.tokens[i+1]['keys'][-1] != t['keys'][-1]):
                if ( (type(obj)==dict) and (obj.has_key('_NEW_'))) and (dumpKey == None) :
                    output+=string.join(obj['_NEW_'],'')
                    del obj['_NEW_']            
        if self.objs.has_key('_NEW_') and (dumpKey == None):
             output += string.join(self.objs['_NEW_'],'')
             del self.objs['_NEW_']
        # kludge to workaround named bug 15195 (named cannot handle multiple keys in rndc.key):
        if self.objs.has_key('key') and self.objs['key'].has_key('_NEW_'):
            output += string.join(self.objs['key']['_NEW_'],'')
            del self.objs['key']['_NEW_']
        if( (dumpKey != None) and len(output) and (output[-1] != '\n')):
            output += '\n';
        return output

    def update(self):
        debug( 'NMDC UPDATE')
        for f in self.files:
            self.tokens=f['tokens']
            self.contents=f['contents']             
            f['contents']=self.out()
        self.objs={}
        self.lists={}        
        for f in self.files:
            self.tokens=[]
            self.contents=f['contents']
            self.parse()
            f['tokens']=self.tokens
            if md5.new(f['contents']).digest() != f['md5']:
                self.dirty = True
        self.tokens=self.files[0]['tokens']
        self.contents=self.files[0]['contents']
        
    def save(self):
        d=datetime.datetime.now()
        ts=str(d.year)+'-'+str(d.month)+'-'+str(d.day)+'_'+str(d.hour)+'.'+str(d.minute)+'.'+str(d.second)
        self.update()
        self.errors=[]
        self.lists={}
        chrootArg=''
        if len(self.chroot):
            chrootArg = '-t '+self.chroot
        for f in self.files:
            md5sum=md5.new(f['contents']).digest()
            if md5sum != f['md5']:
                try:
                    os.rename(f['file'],f['file']+'.'+ts)
                    fd = os.open(f['file'],os.O_CREAT | os.O_WRONLY, 0644)
                    os.write(fd, f['contents'])
                    os.fsync(fd)
                    os.close(fd)
                    commands.getstatusoutput( "if [ -x /sbin/restorecon ]; then /sbin/restorecon "+ f['file'] + "; fi" )
                    os.chown(f['file'],0,25)
                except OSError, e:
                    self.errors = [ e[0], _('cannot write file %s: %s.') % (f['file'], e[1]) ]
                    return e[0]
                f['md5']=md5sum
                fName=f['file']
                if (len(self.chroot)>0) and (len(fName)> len(self.chroot)) and (fName[:len(self.chroot)] == self.chroot):
                    fName = fName[(len(self.chroot)+1):]
                s=commands.getstatusoutput('/usr/sbin/named-checkconf '+chrootArg + ' '+fName)
                if s[0] != 0:
                    self.errors.append(_('Generation of new file "%s" failed:\n%s\nThe original file will be restored.')
                                       % ( f['file'], s[1] )
                                      )
                    try:
                        os.rename(f['file'], f['file']+'.REJECT'+'.'+ts)
                        os.rename(f['file']+'.'+ts,f['file'])
                    except OSError, e:
                        self.errors.append(_('Restoration of file "%s" failed:\n%s\n.')
                                           %(f['file'], str(e[0]), e[1])
                                          )
                        return e[0]
                    return s[0]
        self.dirty = False
        self.tokens=self.files[0]['tokens']
        self.contents=self.files[0]['contents']
        self.lists={}
        return 0

    def preview(self):
        for f in self.files:
            yield ( f['file'], f['contents'] )
        
    def dump(self):
        out=''
        self.lists={}
        includes=None
        if( self.objs.has_key('include')):            
            includes=self.objs['include']
            del self.objs['include']
        for f in self.files:
            self.tokens=f['tokens']
            self.contents=f['contents']
            out += self.out()
        if( includes != None):
            self.objs['include']=includes
        self.tokens=self.files[0]['tokens']
        self.contents=self.files[0]['contents']
        self.lists={}
        return out

    def key_files(s):
        keys=[]
        keyFiles=[]
        if len(s.files) <= 1:
            return keyFiles
        for f in s.files[1:]:
            keyFile = f
            for t in f['tokens']:
                if t['keys'][0] != 'key':
                    keyFile = None
                    keys=[]
                    break
                if (len(keys)==0) or (t['keys'][1] != keys[-1]):
                    keys.append(t['keys'][1])
            if (keyFile != None) and len(keys):
                for k in keys:
                    keyFiles.append([k,f])
        return keyFiles
    
    def firstView(s, name='default', inACL='{ any;};', outACL='{ any;};' ):
        if s.objs.has_key('view'):
            return
        if not s.objs.has_key('zone'):
            return
        fKeys={}
        keyFiles = s.key_files()
        if len(keyFiles):
            for key,f in keyFiles:
                debug( 'KEY FILE:', key, f)
                fKeys[key]=s.objs['key'][key]
                del s.objs['key'][key]
        zones = s.objs['zone']
        del s.objs['zone']
        if s.objs.has_key('include'):
            del s.objs['include']
        contents=s.dump()
        objs={}
        for obj in s.objs.keys():
            if( obj != 'include'):
                objs[obj] = s.objs[obj]
                del s.objs[obj]
        s.objs['zone']=zones
        s.noHeader=True
        contents += '\n\nview "'+name+'" {\n\tmatch-clients '+inACL+'\n\tmatch-destinations '+outACL+'\n\trecursion yes;\n\t' + s.dump()
        s.noHeader=False
        contents += '\n};\n'        
        for k in objs.keys():
            s.objs[k] = objs[k]
        s.files=s.files[:1]
        s.files[0]['contents']=contents
        s.files[0]['tokens']=[]
        s.tokens=[]
        s.objs={}
        s.contents=s.files[0]['contents']
        s.parse()            
        s.files[0]['tokens']=s.tokens
        if len(keyFiles):
            dir = s.chroot
            kfs={}
            if not s.objs.has_key('key'):
                s.objs['key']={}
            for key,file in keyFiles:
                s.objs['key'][key]=fKeys[key]
                if not kfs.has_key(file['file']):
                    f = file['file']
                    if (len(dir) > 0) and (len(f) > len(dir)) and (f[:len(dir)] == dir):
                        f=f[(len(dir)):]                
                    s.files[0]['contents']+='\ninclude "'+f+'"'+";\n"                
                    s.files.append(file)
                    s.contents=s.files[0]['contents']
                    kfs[file['file']]=key
        s.dirty=True
    
    def __init__(self, named_conf=None):
        self.objs={}
        self.contents=''
        self.tokens=[]
        self.files=[]
        self.errors=[]
        self.chroot=''
        self.dirty = False
        self.noHeader=False
        self.lists={}
        e = self.load(named_conf)
#        print self.objs
#        for z in self.objs['zone'].keys():
#           print 'ZONE',z,' file:',self.objs['zone'][z]['file']
#        print self.tokens
#        del self.objs['zone']['"." IN']
#        del self.objs['zone']['"." IN']['allow-update']
#        del self.objs['zone']['"." IN']['allow-update']['any']
#        self.objs['zone']['"." IN']['allow-update']['_NEW_']='none'
#        self.objs['_NEW_']='zone "new.domain.com" IN {\n\ttype master;\n\tfile "new.domain.com";\n};\n'
#        self.objs['options']['_NEW_']='enable-v6-synthesis;\n\t'
#        self.objs['options']['statistics-file']='"/home/jason/new_stats_file_location.txt"'
#        del self.objs['zone']['"0.0.127.in-addr.arpa" IN']['allow-update']
#        os.write(1,self.out())

    def __getattr__(self,name):
        debug( 'NMDC GETATTR:',name)
        if len(name)==0:
            return self.objs
        if self.objs.has_key(name):
            debug( '  GETATTR: objs[',name,']')
            return self.objs[name]
        return None

    def __len__(self):
        return len(self.objs)

    def __contains__(self, keyIn):
        debug( 'NMDC CONTAINS:',keyIn)
        if type(keyIn) ==  list:
            key=keyIn[:]
            obj=self.objs
            while (len(key)>0):
                k=key.pop(0)
                objType = type(obj)
                if objType not in ( dict, list ):
                    break
                if (objType == dict) and obj.has_key(k):
                    obj=obj[k]                
                elif(objType == list) and (len(obj) > k):
                    obj=obj[k]
                else:
                    break
                if( len(key) == 0):
                    debug( ' NMDC CONTAINS:',True)
                    return True
            debug( ' NMDC CONTAINS:',False)
            return False
        else:
            return self.objs.has_key(keyIn)

    def __getitem__(self, keyIn):
        debug( 'NMDC GETITEM',keyIn)
        if type(keyIn) ==  list:
            key=keyIn[:]
            if len(key)==0:
                return self.objs
            obj=self.objs
            while (len(key)>0):
                k=key.pop(0)
                objType = type(obj)
                if objType not in ( dict, list ):
                    break
                elif (objType == dict) and obj.has_key(k):                
                    obj=obj[k]
                elif( objType == list) and (len(obj) > k):
                    obj=obj[k]
                else:
                    break
                if( len(key) == 0):
                    debug( '  NMDC GET OK',k )
                    return obj
            raise IndexError
        elif self.objs.has_key(keyIn):
            debug( '  NMDC GET OK',keyIn )
            return self.objs[keyIn];
        elif len(keyIn) == 0:
            debug( '  NMDC GET objs' )
            return self.objs
        else:
            raise IndexError
    
    def __delitem__(self, keyIn):
        debug( 'NAMED CONF DELETE',keyIn)
        if type(keyIn) ==  list:
            obj=self.objs
            key=keyIn[:]
            k=''
            while (len(key)>0):
                k=key.pop(0)
                objType = type(obj)
                if objType not in ( dict, list ):
                    break
                if (objType == dict) and not obj.has_key(k):
                    debug( ' NMDC DEL: bad dict key:',k)
                    break
                elif(objType == list) and (len(obj) <= k):
                    debug( ' NMDC DEL: bad list key:',k)
                    break
                if( len(key) == 0):
                    debug( ' NMDC DEL OK [',k,']')
                    del obj[k]
                    return
                obj = obj[k]                
        elif self.objs.has_key(keyIn):
            debug( ' NMDC DEL OK ',keyIn)
            del self.objs[keyIn];
            return
        raise IndexError
        
    def __setitem__(self,keyIn,value):
        debug( 'NAMED CONF SET',keyIn)
        if type(keyIn) ==  list:
            obj=self.objs
            key=keyIn[:]
            while (len(key)>0):
                k=key.pop(0)
                objType = type(obj)
                if objType not in ( dict, list ):
                    break
                if (objType == dict) and not obj.has_key(k):
                    break
                elif(objType == list) and (len(obj) <= k):
                    break
                if len(key)==0:
                    debug( ' NMDC SET OK [',k,']' )
                    obj[k] = value
                    return
                obj=obj[k]
            if( len(key) ):
                raise IndexError
                return
        elif self.objs.has_key(keyIn):
            debug( ' NMDC SET OK',k )
            self.objs[keyIn] = value
            return
        else:
            obj=self.objs
        if type(obj) == dict:
            if obj.has_key('_NEW_'):
                if type(obj['_NEW_'])!=list:
                    obj['_NEW_']=[obj['_NEW_']]
                debug( ' NMDC NEW APPEND OK')
                obj['_NEW_'].append(value)
            else:
                obj['_NEW_'] = value
                debug( ' NMDC NEW OK', obj                )
        elif type(obj) == list:
            debug( ' NMDC NEW LIST APPEND OK')
            obj.append({'_NEW_':value})
        else:
            raise IndexError

    def __deepcopy__(s,memo):
        return s

if __name__ == "__main__":
    import __builtin__
    import gettext
    __builtin__.__dict__['SCB']='/usr/share/system-config-bind'
    os.environ['SCB']='/usr/share/system-config-bind'
    os.putenv('SCB','/usr/share/system-config-bind')
    gettext.bindtextdomain('bindconf', '/usr/share/locale')
    gettext.textdomain('bindconf')
    __builtin__.__dict__['_'] = gettext.gettext    
    import Debug
    dbg = os.getenv('SCB_DEBUG')
    if dbg != None:
        Debug.DEBUG_CWD = SCB
        Debug.DEBUG = 1
        Debug.parseEnv(dbg,Debug)
    nmdc=NamedConf()
    if len(nmdc.errors):
        print  "Errors: ",nmdc.errors
        sys.exit(-1)
#    nmdc['options']='options {\n};\n\n'
#    nmdc.update();
#    print nmdc['options'];
#    nmdc.objs['options']['_NEW_']='directory "/var/named";\n'
#    print nmdc['options'];
#    nmdc.update()
#    print nmdc.contents    
#    print nmdc['options'];
#    os.write(1,nmdc.out())
#    nmdc.firstView("default", "match-clients { any; };\n", "match-destinations { any ;}\n")
#    os.write(1,nmdc.out())
#    dumpKey = ['view', '"default"', 'zone', '"0.in-addr.arpa." IN' ];    
#    del nmdc['trusted-keys']
#    nmdc['trusted-keys']='trusted-keys { k. 1 2 3 key; k. 4 5 6 key; };\n'
#    nmdc.update()
#    print nmdc['trusted-keys']
#    os.write(1,nmdc.out())
#    os.write(1,nmdc.out())
#    print nmdc.tokens
#    print nmdc.objs['trusted-keys']
#    print nmdc.objs['acl']
#    print nmdc.options
#    print nmdc.acl
#    del nmdc.acl['test']['!fec0:ac10:40a5:1::/64']
#    del nmdc.acl['test']['!172.16.83/22']
#    print nmdc.acl
#    for origin, real_key, file, clas, zone, view in nmdc.zones():
#        print origin, file, clas, view
#    del nmdc.acl['test']['!']
#    nmdc[ 'options' ]['_NEW_']='\tallow-recursion { any; };\n'
#    nmdc[['zone','newdomain']]='zone "newdomain." IN {\n\tfile "newdomain.zone";\n\ttype master;\n};\n'
#    del nmdc[['zone','"." IN','allow-update']]
#    nmdc[['zone','"." IN','allow-update']]='allow-update { none; };\n\t'
#    nmdc[['options','recursion']]='recursion yes;\n\t'
#    nmdc[['zone', '"localhost." IN','recursion']]='recursion no;\n\t'
#    print nmdc.zone
#    print nmdc['zone']['"." IN']
#    print nmdc[''].keys()
#    print nmdc['zone'].keys()
#    del nmdc.zone['"." IN']
#    print nmdc[['view','"default"','zone','"." IN']]
#    del nmdc[['view','"default"','zone','"." IN','allow-update']]
#    del nmdc.zone['"." IN']['zone-statistics']
#    nmdc[ ['view','"default"','zone', '"." IN', 'allow-update'] ] = 'allow-update {any;};\n'
#    del nmdc.zone['"test2." IN']
#    else:
#        os.write(1,nmdc.out())
#        print nmdc['']
#        print nmdc.objs
#        nmdc.update()
#        nmdc.dump()
#        if len(nmdc.errors):
#            print "Error: ",nmdc.errors
#    nmdc[['options','enable-v6-synthesis']]='enable-v6-synthesis { any;};\n\t'
#    z=nmdc.dump()
#    del nmdc.objs['zone']
#    nmdc.firstView()
#    os.write(1,nmdc.out())
#    for origin, real_key, filename, clas, cfg, view in nmdc.zones():
#        print  origin, real_key, filename, clas, cfg, view
#    print nmdc.acl['test']
#    print nmdc[['acl', 'test', '', '', '', 1]]
#    del nmdc[['acl', 'test', '', '', '', 1]]
#    nmdc[['acl', 'test', '', '', '', 1]]='{ 172.16.83.127/25; 172.16.83.128/25; };'
#    os.write(1,nmdc.out())
#    print nmdc[['acl','test']]
#    del nmdc[['acl','test']]
#    nmdc[['acl','test']]='acl test { any; };\n'
#    nmdc[['view','"new"','recursion']]='recursion yes;\n'
#    print nmdc[['view','"new"']]
#    os.write(1,nmdc.out())
#    print nmdc.files
#    del nmdc [['options','allow-query']]
#    nmdc [['acl','redhat']]='acl redhat { 172.16/16; 192.168/16; localnets; };\n'
#    nmdc [['options','allow-query']] = 'allow-query { redhat; };\n'
#    print nmdc.tokens[0]
#    nmdc.update()
#    print nmdc.contents
#    print nmdc.tokens[0]
#    print nmdc['options']
#    for t in nmdc.tokens:
#        if t['keys'] == ['options','query-source']:
#            print t
#    nmdc[['options','query-source']]='address 172.16.80.40 port 53'
#    print nmdc['options']    
#    os.write(1,nmdc.out())
#    os.write(1,nmdc.out(['view','"default"','zone','"tst." IN']))
#    print nmdc[[
#    del nmdc[['view','"default"','zone','"redhat.com" IN', 'forwarders']]
#    del nmdc[['view','"default"','zone','"." IN','allow-update']]
#    os.write(1,nmdc.out())
#    print nmdc.out(['controls'])
#    print nmdc['controls']
#    print nmdc['key','rndckey']
#    os.write(1,nmdc.out())
#    print nmdc[['options','sortlist']]
